// Teensy_3.1_Chess_Program_Update
// Written by Theron Wierenga, October 2016

// Include the Neopixel library to drive RGB LEDs
#include <Adafruit_NeoPixel.h>
#include <avr/power.h>
#define NUMPIXELS  256  // Number of LEDs
#define NEOPIN 11       // Teensy 3.1 pin for NEOPIXELS
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, NEOPIN, NEO_GRB + NEO_KHZ800);
// Wire Library is used for I2C communication with the LCD display
#include <Wire.h>
// Include the LCD library
#include <LiquidCrystal_I2C.h>
// Create an LCD object
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  // Set the LCD I2C address for 20x4 LCD

// Defines for colors
#define BLACK 0
#define RED 1
#define ORANGE 2
#define YELLOW 3
#define GREEN 4
#define CYAN 5
#define BLUE 6
#define VIOLET 7
#define MAGENTA 8
#define WHITE 128

// Misc defines
#define SWITCH A3
#define LEDPIN 13
#define RETRIEVE 0
#define STORE 1

// Defines for chess pieces and spaces
#define BORDER 255
#define EMPTY 0
#define BLACK_PAWN 1
#define BLACK_KNIGHT 3
#define BLACK_BISHOP 4
#define BLACK_ROOK 5
#define BLACK_QUEEN 9
#define BLACK_KING 16
#define WHITE_PAWN 129
#define WHITE_KNIGHT 131
#define WHITE_BISHOP 132
#define WHITE_ROOK 133
#define WHITE_QUEEN 137
#define WHITE_KING 144

//byte piecesVal[12][12] =  {{255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
//                           {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
//                           {255, 255,   5,   3,   4,   9,  16,  4,   3,   5, 255, 255},    // White = 128, white at bottom, starting positions
//                           {255, 255,   1,   1,   1,   1,   1,   1,   1,   1, 255, 255},    // Black =   0, add color to piece value
//                           {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},    //
//                           {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},    //       Black   White
//                           {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},    // Pawn =    1 and 129
//                           {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},    // Knight =  3     131
//                           {255, 255, 129, 129, 129, 129, 129, 129, 129, 129, 255, 255},    // Bishop =  4     132
//                           {255, 255, 133, 131, 132, 137, 144, 132, 131, 133, 255, 255},    // Rook =    5     133
//                           {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},    // Queen =   9     137
//                           {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255}};   // King =   16     144

// This array is the equivalent of the one above that is commented out
byte piecesVal[12][12] =  {{BORDER, BORDER, BORDER,     BORDER,       BORDER,       BORDER,      BORDER,     BORDER,       BORDER,       BORDER,     BORDER, BORDER},
                           {BORDER, BORDER, BORDER,     BORDER,       BORDER,       BORDER,      BORDER,     BORDER,       BORDER,       BORDER,     BORDER, BORDER},
                           {BORDER, BORDER, BLACK_ROOK, BLACK_KNIGHT, BLACK_BISHOP, BLACK_QUEEN, BLACK_KING, BLACK_BISHOP, BLACK_KNIGHT, BLACK_ROOK, BORDER, BORDER},
                           {BORDER, BORDER, BLACK_PAWN, BLACK_PAWN,   BLACK_PAWN,   BLACK_PAWN,  BLACK_PAWN, BLACK_PAWN,   BLACK_PAWN,   BLACK_PAWN, BORDER, BORDER},
                           {BORDER, BORDER, EMPTY,      EMPTY,        EMPTY,        EMPTY,       EMPTY,      EMPTY,        EMPTY,        EMPTY,      BORDER, BORDER},
                           {BORDER, BORDER, EMPTY,      EMPTY,        EMPTY,        EMPTY,       EMPTY,      EMPTY,        EMPTY,        EMPTY,      BORDER, BORDER},
                           {BORDER, BORDER, EMPTY,      EMPTY,        EMPTY,        EMPTY,       EMPTY,      EMPTY,        EMPTY,        EMPTY,      BORDER, BORDER},
                           {BORDER, BORDER, EMPTY,      EMPTY,        EMPTY,        EMPTY,       EMPTY,      EMPTY,        EMPTY,        EMPTY,      BORDER, BORDER},
                           {BORDER, BORDER, WHITE_PAWN, WHITE_PAWN,   WHITE_PAWN,   WHITE_PAWN,  WHITE_PAWN, WHITE_PAWN,   WHITE_PAWN,   WHITE_PAWN, BORDER, BORDER},
                           {BORDER, BORDER, WHITE_ROOK, WHITE_KNIGHT, WHITE_BISHOP, WHITE_QUEEN, WHITE_KING, WHITE_BISHOP, WHITE_KNIGHT, WHITE_ROOK, BORDER, BORDER},
                           {BORDER, BORDER, BORDER,     BORDER,       BORDER,       BORDER,      BORDER,     BORDER,       BORDER,       BORDER,     BORDER, BORDER},
                           {BORDER, BORDER, BORDER,     BORDER,       BORDER,       BORDER,      BORDER,     BORDER,       BORDER,       BORDER,     BORDER, BORDER}};
                            
byte piecesValCur[12][12] = {{255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},  // Current piece values
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
                             {255, 255,   5,   3,   4,   9,  16,   4,   3,   5, 255, 255},
                             {255, 255,   1,   1,   1,   1,   1,   1,   1,   1, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255, 129, 129, 129, 129, 129, 129, 129, 129, 255, 255},
                             {255, 255, 133, 131, 132, 137, 144, 132, 131, 133, 255, 255},
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255}};

byte piecesValHist[12][12] = {{255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255}, // Previous board before last move
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
                             {255, 255,   5,   3,   4,   9,  16,   4,   3,   5, 255, 255},
                             {255, 255,   1,   1,   1,   1,   1,   1,   1,   1, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255,   0,   0,   0,   0,   0,   0,   0,   0, 255, 255},
                             {255, 255, 129, 129, 129, 129, 129, 129, 129, 129, 255, 255},
                             {255, 255, 133, 131, 132, 137, 144, 132, 131, 133, 255, 255},
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255},
                             {255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255}};
                           
byte pieces[8][8] =        {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x] Starting positions
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesCurrent[8][8] = {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x] Where current piece are located
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesHist[8][8] =    {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x] Previous board before last move
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesTemp1[8][8] =   {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x]  // Used to read pieces looking for changes
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesTemp2[8][8] =   {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x]  // Used to read pieces looking for changes
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesTemp3[8][8] =   {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x]  // Used to read pieces looking for changes
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

byte piecesError[8][8] =   {{1, 1, 1, 1, 1, 1, 1, 1}, // [y][x]  // Used to read pieces after an error
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {0, 0, 0, 0, 0, 0, 0, 0},
                            {1, 1, 1, 1, 1, 1, 1, 1},
                            {1, 1, 1, 1, 1, 1, 1, 1}};

// Characters to produce standard chess notation
// x = columns, files, a in LL, a b c d e f g h
// y = rows, ranks, 1 in LL, 1 2 3 4 5 6 7 8 
// The String rankFile(int x, int y) function will use these
String files[8] = {"a","b","c","d","e","f","g","h"};
String ranks[8] = {"8","7","6","5","4","3","2","1"};

// Create an array of 9 standard colors
// black, red, orange, yellow, green, cyan, blue, violet, magenta
uint32_t colors[9] = { pixels.Color(0, 0, 0),     pixels.Color(255, 0, 0),   pixels.Color(255, 60, 0),
                       pixels.Color(255, 160, 0), pixels.Color(0, 255, 0),   pixels.Color(0, 255, 255), 
                       pixels.Color(0, 0, 255),   pixels.Color(148, 0, 211), pixels.Color(255, 0, 255)};
                       
int c;                                                    // For returned count
bool same, flag;                                          // To compare arrays and flag end of loops
byte turn = WHITE;                                        // White's turn first
byte bdCount = 32, bdCountHist;                           // 32 pieces to start, number of pieces currently in play
int pathX[30], pathY[30], pathVal[30], pathCount = 0;     // Path values and total paths
int xx1, xx2, xx3, yy1, yy2, yy3;                         // Active piece locations
int xKingCheck, yKingCheck, xAttackPos, yAttackPos;       // Stored positions for king in check and its attacking piece
int xPathtoKing[6], yPathtoKing[6], pathtoKingCount;      // Path between king and attacking piece
String message;                                           // For LCD output

// Functions

// Main logic loop
// int mainLoop(byte turn)

// Compare piece arrays
// bool comparePieceArrays(int &xx, int &yy, byte piecesCurrent[][8], byte piecesTemp1[][8])

// Is position on a path?
// bool onPath(int xx, int yy)

// Toggles turn for certain use of getPaths
// int nextTurn(int turn)

// Is the King in check ?
// bool kingInCheck(int turn)

// Are we in checkmate ? return true if in checkmate
// bool checkMate(int turn)

// If king moves to this path will it be in check, ie. moves into path of opponent
// bool squareInOpponentPath(int xx, int yy, int turn)

// Get path from king to attacking piece
// void pathToKingInCheck(int xKingCheck, int yKingCheck, int xAttackPos, int yAttackPos)

// Follow a single path to its end
// void getSinglePathinPaths(int xx, int yy, int xdir, int ydir, int &count, int turn)

// Get the legal positions a piece can move to
// int getPaths(byte val, int x, int y, int turn)

// Color the positions a piece can legally move to
// void colorPath(int count, int turn)

// Read a row of 8 Hall Effect sensors
// int readLine(int row, byte piecesTemp[][8])

// Read all 8 lines of the Hall Effect Sensors
// int readHall(byte piecesTemp[][8])

// Store or retrieve prievous board settings to be able to undo a move
// void store_retrieveHist(int dir, byte piecesValCur[][12], byte piecesValHist[][12], byte piecesCurrent[][8], byte piecesHist[][8], byte bdCount, byte bdCountHist)

// Color an x,y square with a specific color
// void colorSquare(byte x, byte y, uint32_t color, bool on)

// Color all squares, red for error alert, blue for checkmate
// Seems to work OK without a delay
// void colorBoard(uint32_t color)

// Color a chessboard with blue (for black) and red squares
// void standBoard()

// Show an error state and wait for correction, ie. all pieces replaced to positions before the error
// void showError(byte piecesCurrent[][8], byte piecesError[][8])

// The name of the piece for LCD readout and debugging
// String namePiece(int x, int y)

// print the piece locations to the Serial monitor for debugging
// void printPiecesLoc(byte pieces[][8])

// print the piece values to the Serial monitor for debugging
// void printPiecesVal(byte pieces[][12])

// Print the paths to the Serial monitor for debugging
// void printPaths(int x, int y)

// myDebug, print two numbers and a string to the Serial monitor for debugging
// void myDebug(int x, int y, String mess)

// Return the traditional rank and file notation denoting the position of a square, so position 3,4 equals d5
// This is used for output to the LCD display
// String rankFile(int x, int y)


// *************************************************************************************************************************
void setup() {

  Serial.begin(115200);       // Activate serial ouput to serial monitor
  pinMode(2, INPUT_PULLUP);   // 0  These lines read the columns
  pinMode(3, INPUT_PULLUP);   // 1
  pinMode(4, INPUT_PULLUP);   // 2
  pinMode(5, INPUT_PULLUP);   // 3
  pinMode(6, INPUT_PULLUP);   // 4
  pinMode(7, INPUT_PULLUP);   // 5
  pinMode(8, INPUT_PULLUP);   // 6
  pinMode(12, INPUT_PULLUP);  // 7
  pinMode(9, INPUT);          // Reserved for Serial2 RX2
  pinMode(10, OUTPUT);        // Reserved for Serial2 TX2
  pinMode(11, OUTPUT);        // Drives the RGB LED strip
  pinMode(13, OUTPUT);        // LED on Teensy 3.1, ON WHITE turn, OFF BLACK turn
  pinMode(A0, OUTPUT);        // 3 bit output to 74138
  pinMode(A1, OUTPUT);        // 3 bit output to 74138
  pinMode(A2, OUTPUT);        // 3 bit output to 74138
  pinMode(A3, INPUT_PULLUP);  // Switch input for selection or restart

  pixels.begin();             // This initializes the NeoPixel library.
  pixels.setBrightness(8);    // Brightness can be set from 0 - 255.

  digitalWrite(A0, LOW);      // A0 output to 74138 - Turn off all output lines on startup
  digitalWrite(A1, LOW);      // A1 output to 74138
  digitalWrite(A2, LOW);      // A1 output to 74138
  digitalWrite(9, LOW);       // Reserved for Serial2 RX2
  digitalWrite(10, LOW);      // Reserved for Serial2 TX2
  digitalWrite(11, LOW);      // Drives the RGB LED strip
  digitalWrite(13, HIGH);     // WHITE turn to start
  colorBoard(colors[RED]);                 // Turn on all LEDs to red
  lcd.begin(20, 4);           // Setup the LCD display
  lcd.setCursor(0, 0);
  lcd.print("Chessboard Version 2");
  lcd.setCursor(0, 1);
  lcd.print(" White moves first");
  delay(1000);
  do                                    // Check to make sure all chess pieces are in starting positions
   {
     c = readHall(piecesTemp1);         // read the Hall Effect sensors
     Serial.println(c);
     same = comparePieceArrays(xx1, yy1, pieces, piecesTemp1);
     printPiecesLoc(piecesTemp1);
   }while ((c != bdCount) || (!same));  // wait until all pieces in correct places  
   standBoard();                        // Display the standard board, RED = red squares, BLUE = black squares 
}

// *************************************************************************************************************************
void loop() {

  // If a restart switch between A3 and ground is not closed, this code block will have no effect
  byte restart = digitalRead(SWITCH);
  if (restart == 0)                      // Restart switch on. new game
  {
    for (int y = 0; y < 8; y++)
    {
      for (int x = 0; x < 8; x++)       // Return all arrays to starting values
      {
        piecesValCur[y + 2][x + 2] = piecesVal[y + 2][x + 2];
        piecesCurrent[y][x] = pieces[y][x];
      }
    }
    turn = WHITE;
    bdCount = 32;
    do                                   // Check to make sure all chess pieces are in starting positions
    {
      c = readHall(piecesTemp1);         // read the Hall Effect sensors
      Serial.println(c);
      same = comparePieceArrays(xx1, yy1, pieces, piecesTemp1); 
    }while ((c != bdCount) || (!same));  // wait until all pieces in correct places  
    standBoard();
  }
  
// *************************************************************************************************************************

  mainLoop(turn);                            // The main core of the program, checks for moves, calls errors
  flag = true;                               // Set a flag for the king in check test below
  if (kingInCheck(turn)) 
  {
    // If player moved a piece so their king is in check, retrieve previous values and throw and error
    store_retrieveHist(RETRIEVE, piecesValCur, piecesValHist, piecesCurrent, piecesHist, bdCount, bdCountHist);
    showError(piecesCurrent, piecesError);                        // moved into check
    colorSquare(xAttackPos, yAttackPos, colors[VIOLET], true);    // recolor the position of the attacking piece
    flag = false;
  }
  if (flag)                                  // King not in check, so next turn
  {
    if (turn == WHITE)                      
    {
      turn = BLACK;
      digitalWrite(LEDPIN, LOW);             // Not needed with LCD output ?
    }
    else
    {
      turn = WHITE;
      digitalWrite(LEDPIN, HIGH);
    }
    if (kingInCheck(turn))                   // This test for king in check reveals a move placing oponents king in check
    {
      // Check now for checkmate
      if (checkMate(turn))
      {
        colorBoard(colors[BLUE]);
        while(true) {}                      // loop forever
      }
    }
  }
}

// *************************************************************************************************************************
// Main logic loop
int mainLoop(byte turn)
{
  bool flag;                                // Used for ending the do loop
  int total1 = 0, total2 = 0, total3 = 0;   // For different reads of total number of pieces on the board
  do
  {
    flag = false;
    do
    {
      total1 = readHall(piecesTemp1);       // Read the Hall Effect sensors
    }while (total1 == bdCount);             // Wait for a change
    comparePieceArrays(xx1, yy1, piecesCurrent, piecesTemp1);     //xx1, yy1 is the space with piece removed
    if (((turn == WHITE) && (piecesValCur[yy1 + 2][xx1 + 2] < WHITE_PAWN)) ||  // Make sure right color takes turn
        ((turn == BLACK) && (piecesValCur[yy1 + 2][xx1 + 2] > BLACK_KING)))
    {
      showError(piecesCurrent, piecesError);   // Throw error is not right colorn
      flag = true;
      continue;
    }
    myDebug(xx1, yy1, "Piece removed at");
    lcd.clear();
    lcd.setCursor(0, 0);
    message = namePiece(xx1, yy1) + rankFile(xx1, yy1);
    lcd.print(message);
    pathCount = 0;
    int paths = getPaths(piecesValCur[yy1 + 2][xx1 + 2], xx1, yy1, turn); // Get all legal paths (squares) 
                                                                          // If no paths, no green or yellow squares
    colorPath(paths, turn);
   // printPaths(xx1, yy1);   // Can be used for debugging
    do
    {
      total2 = readHall(piecesTemp2);            // Read the Hall Effect sensors
    }while (total2 == total1);                   // Wait for a change

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    if (bdCount - total2 == 2)                   // Second piece removed - so this should be a capture
    {
      comparePieceArrays(xx2, yy2, piecesTemp2, piecesTemp1);     // xx2, yy2 will be position of 2nd piece removed
      for (int i=0; i < pathCount; i++)
      {
        if ((pathX[i] == xx2) && (pathY[i] == yy2))               // found the 2nd piece
        {
          myDebug(xx2, yy2, "Second piece removed at");           // More debugging output
          Serial.println(namePiece(xx2, yy2));
          do
          {
            total3 = readHall(piecesTemp3);                       // Read the Hall Effect sensors
          }while (total2 == total3);                              // Wait for a change, the piece show be
                                                                  // replaced on the captured piece's square
          comparePieceArrays(xx3, yy3, piecesTemp3, piecesTemp2);
          if ((xx2 == xx3) && (yy2 == yy3))                       // Captured piece and replaced piece
          {                                                       // square should be the same
            // save previous board positions and values
            store_retrieveHist(STORE, piecesValCur, piecesValHist, piecesCurrent, piecesHist, bdCount, bdCountHist);
            lcd.setCursor(0, 1);
            lcd.print(" to ");
            lcd.setCursor(4, 1);
            message = rankFile(xx2, yy2);
            lcd.print(message);
            lcd.setCursor(0, 2);
            message = namePiece(xx2, yy2) + rankFile(xx2, yy2);
            lcd.print(message);
            lcd.setCursor(0, 3);
            lcd.print(" captured");
            piecesCurrent[yy1][xx1] = 0;
            piecesCurrent[yy2][xx2] = 1;                          // actually piece already there
            piecesValCur[yy2 + 2][xx2 + 2] = piecesValCur[yy1 + 2][xx1 + 2];    
            piecesValCur[yy1 + 2][xx1 + 2] = 0;
            bdCount = bdCount - 1;
            flag = true;
            //printPiecesVal(piecesValCur);
            continue;
          }
        }
      }
      if (flag)             // 2nd piece was captured and removed
      {
        flag = false;       // this will end our do loop
        continue;
      }
      else
      {
        // Flag wasn't set to true above so captured piece and replaced piece 
        // in different locations, an error
        showError(piecesCurrent, piecesError);
      }
    }

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    if (comparePieceArrays(xx2, yy2, piecesCurrent, piecesTemp2) == true) // arrays are the same
    {
      // Picked up piece was just returned to to original location - no move
      flag = true;          // Continue the do loop
      standBoard();         // Display the standard board, erasing any paths
      lcd.setCursor(0, 1);
      lcd.print(" Piece returned");
      continue;
    }

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     // This section processes a simple move with no capture
     comparePieceArrays(xx2, yy2, piecesTemp1, piecesTemp2);
     myDebug(xx2, yy2, "Piece down at");
     if (!onPath(xx2, yy2))                     // Check to make sure piece was place on an allowable square
     {
        showError(piecesCurrent, piecesError);
        flag = true;
        continue;
     }
     // Save previous board positions and values, this is for possible king in check detected later 
     store_retrieveHist(STORE, piecesValCur, piecesValHist, piecesCurrent, piecesHist, bdCount, bdCountHist);
     // If we get here piece was moved from xx1, yy1 to xx2, yy2
     // Write the new values into piecesCurrent and piecesValCur arrays
     piecesCurrent[yy1][xx1] = 0;
     piecesCurrent[yy2][xx2] = 1;
     piecesValCur[yy2 + 2][xx2 + 2] = piecesValCur[yy1 + 2][xx1 + 2];
     piecesValCur[yy1 + 2][xx1 + 2] = 0;
     lcd.setCursor(0, 1);
     lcd.print(" to ");
     lcd.setCursor(4, 1);
     message = rankFile(xx2, yy2);
     lcd.print(message);
  } while (flag); 
  pathCount = 0;      // Reset the pathCount
  standBoard();       // Display the standard board, erasing any paths
  Serial.println();
  Serial.println("Through mainLoop");
  Serial.println();
  return 0;
}

// *************************************************************************************************************************
// Compare piece arrays
bool comparePieceArrays(int &xx, int &yy, byte piecesCurrent[][8], byte piecesTemp1[][8])
{
  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      if (piecesTemp1[y][x] != piecesCurrent[y][x])
      {
        // First piece that doesn't match previous,
        // Write to xx, yy and return
        xx = x; yy = y;
        return false;         // arrays are different
      }
    }
  }
  return true;                // arrays are the same
}

// *************************************************************************************************************************
// Is position on a path?
bool onPath(int xx, int yy)
{
  for (int i = 0; i < pathCount; i++)
  {
    if ((xx == pathX[i]) && (yy == pathY[i]))
    {
      return true;    // When we find position on path return
    }
  }
  return false;       // Position not on path
}

// *************************************************************************************************************************
// Toggles turn for certain use of getPaths
int nextTurn(int turn)
{
  // This just return the opposite of turn, ie. the next turn
  int t;
  if (turn == BLACK)
  {
    t = WHITE;
  }
  else
  {
    t = BLACK;
  }
  return t;
}

// *************************************************************************************************************************
// Is the King in check ?
bool kingInCheck(int turn)
{
  // This one is tricky, we go though entire array looking for the next turn's pieces
  // Then we find the paths for the next turn's pieces and if the turn's king
  // is on one of these paths then that king will be in check. We color the square
  // with the piece that puts the king in check violet.
  pathCount = 0;
  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      if (((turn == WHITE) && (piecesValCur[y + 2][x + 2] >= BLACK_PAWN) && (piecesValCur[y + 2][x + 2] <= BLACK_KING)) ||
          ((turn == BLACK) && (piecesValCur[y + 2][x + 2] >= WHITE_PAWN) && (piecesValCur[y + 2][x + 2] <= WHITE_KING)))
      {
       // Serial.println(piecesValCur[y + 2][x + 2]);
       // myDebug(x, y, "Square");
        getPaths(piecesValCur[y + 2][x + 2], x, y, nextTurn(turn));
        // printPaths(x, y);
        for (int i = 0; i < pathCount; i++)
        {
           if (((pathVal[i] == BLACK_KING) && (turn == BLACK)) || ((pathVal[i] == WHITE_KING) && (turn == WHITE)))
           {
              lcd.setCursor(0, 4);
              if (turn == BLACK)
              {
                Serial.println("Black King in Check");
                lcd.print("Black King in Check");
              }
              else 
              {
                Serial.println("White King in Check");
                lcd.print("White King in Check");
              }
              myDebug(pathX[i], pathY[i], namePiece(pathX[i], pathY[i]));
              // Save the position of the king in check
              xKingCheck = pathX[i]; yKingCheck = pathY[i];
              // Save the position of the attacking piece
              myDebug(x, y, "Attacking piece");
              xAttackPos = x; yAttackPos = y;
              pathToKingInCheck(xKingCheck, yKingCheck, xAttackPos, yAttackPos);
              colorSquare(xAttackPos, yAttackPos, colors[VIOLET], true);
              colorSquare(xKingCheck, yKingCheck, colors[VIOLET], true);
              pathCount = 0;
              return true;
           }
        }
        pathCount = 0;
      }
    }
  }
  return false;
}

// ***************************************************************************************************************************
// Are we in checkmate ? return true if in checkmate
bool checkMate(int turn)
{
  // Function kingInCheck has stored king position in xKingCheck, yKingCheck, and also stored 
  // xAttackPos and yAttackPos of attacking piece.
  // pathToKingInCheck(xKingCheck, yKingCheck, xAttackPos, yAttackPos) stores the path between
  // the two in int variables xPathtoKing[6], yPathtoKing[6], and pathtoKingCount

  // can the king move to get out of check
  // get the paths for the king which is in check
  getPaths(piecesValCur[yKingCheck + 2][xKingCheck + 2], xKingCheck, yKingCheck, turn);
  printPaths(xKingCheck, yKingCheck);
  // pathtoKingCount needs to be zeroed before returning this function
  for (int i = 0; i < pathCount; i++)        // paths king can move to
  {
    for (int j = 0; j < pathtoKingCount; j++)  // attack path to king
    {
      if (!((pathX[i] == xPathtoKing[j]) && (pathY[i] == yPathtoKing[j])))
      {
        // There is a square the king can move to that is not on the path
        // from the attacking piece to the king, but will king move into check from another piece?
        myDebug(pathX[i], pathY[i], "Path king can move to");
        if (squareInOpponentPath(pathX[i], pathY[i], turn))
        {
          // this square will have king in check if king moves there 
          continue;
        }
        else
        {
          // square is open - no checkmate
         
          pathtoKingCount = 0;
          pathCount = 0;
          return false;
        }
      }
    }
  }

  // now we check to see if we can move a piece to block attacking piece

  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      if (((turn == BLACK) && (piecesValCur[y + 2][x + 2] >= BLACK_PAWN) && (piecesValCur[y + 2][x + 2] <= BLACK_QUEEN)) ||
          ((turn == WHITE) && (piecesValCur[y + 2][x + 2] >= WHITE_PAWN) && (piecesValCur[y + 2][x + 2] <= WHITE_QUEEN)))
      {
        pathCount = 0;
        getPaths(piecesValCur[y + 2][x + 2], x, y, turn);
        // printPaths(x, y);
        for (int i = 0; i < pathCount; i++)
        {
          for (int j = 0; j < pathtoKingCount; j++)
          {
            if (((pathX[i] == xPathtoKing[j]) && (pathY[i] == yPathtoKing[j])) ||   // piece can move to attacking path
                ((pathX[i] == xAttackPos) && (pathY[i] == yAttackPos)))             // or can capture the attacking piece
            {
              // we can block attacking piece
              myDebug(pathX[i], pathY[i], "Blocking piece for king");
              pathtoKingCount = 0;
              pathCount = 0;
              return false;
            }
          }
        }
      }
    }
  }
  pathtoKingCount = 0;
  // if we get here, checkmate
  lcd.setCursor(0, 4);
  lcd.print("    CHECKMATE!      ");
  return true;
}

// ***************************************************************************************************************************
// If king moves to this path will it be in check, ie. moves into path of opponent
bool squareInOpponentPath(int xx, int yy, int turn)
{
  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      if (((turn == WHITE) && (piecesValCur[y + 2][x + 2] >= BLACK_PAWN) && (piecesValCur[y + 2][x + 2] <= BLACK_QUEEN)) ||
          ((turn == BLACK) && (piecesValCur[y + 2][x + 2] >= WHITE_PAWN) && (piecesValCur[y + 2][x + 2] <= WHITE_QUEEN)))
      {
        getPaths(piecesValCur[y + 2][x + 2], x, y, nextTurn(turn));
        for (int i = 0; i < pathCount; i++)
        {
          if ((pathX[i] == xx) && (pathY[i] == yy))
          {
            return true;
          }
        }
      }
    }
  }
  return false;
}

// ***************************************************************************************************************************
// Get path from king to attacking piece
void pathToKingInCheck(int xKingCheck, int yKingCheck, int xAttackPos, int yAttackPos)
{
  int xPath, yPath, xMove, yMove, count = 0;
  switch(piecesValCur[yAttackPos + 2][xAttackPos + 2])
  {
    // make sure it's the right find of piece to create a path to king
    case BLACK_BISHOP: break;
    case BLACK_ROOK:   break;
    case BLACK_QUEEN:  break;
    case WHITE_BISHOP: break;
    case WHITE_ROOK:   break;
    case WHITE_QUEEN:  break;
    default:  pathtoKingCount = 0;
              return;
              break;
  }
  if (xKingCheck == xAttackPos)
  {
    xMove = 0;
  }
  else if (xKingCheck < xAttackPos)
  {
    xMove = 1;
  }
  else
  {
    xMove = -1;
  }

  if (yKingCheck == yAttackPos)
  {
    yMove = 0;
  }
  else if (yKingCheck < yAttackPos)
  {
    yMove = 1;
  }
  else
  {
    yMove = -1;
  }
  
  xPath = xKingCheck + xMove;
  yPath = yKingCheck + yMove;
  while ((xPath != xAttackPos) && (yPath != yAttackPos))
  {
    xPathtoKing[count] = xPath;
    yPathtoKing[count] = yPath;
    myDebug(xPathtoKing[count], yPathtoKing[count], "Along the path to King");
    count++;
    xPath += xMove;
    yPath += yMove;
  }
  pathtoKingCount = count;
}

// *************************************************************************************************************************
// Follow a single path to its end
void getSinglePathinPaths(int xx, int yy, int xdir, int ydir, int &count, int turn)
{
  // This function starts with the xx, yy position of a square and xdir and ydir directions to add to the square's
  // position with every loop. The loop continues as long as the squares are EMPTY. When it hits a square occupied
  // it checks to be sure it is an opponent's piece and then will add this to the path.
  xx = xx + xdir;
  yy = yy + ydir;
  while (piecesValCur[yy][xx] == EMPTY)
  {
    pathY[count] = yy - 2; pathX[count] = xx - 2; pathVal[count] = piecesValCur[yy][xx];
    count++;
    xx = xx + xdir;
    yy = yy + ydir;
  }
  if (((turn == WHITE) && (piecesValCur[yy][xx] >= BLACK_PAWN) && (piecesValCur[yy][xx] <= BLACK_KING)) ||
      ((turn == BLACK) && (piecesValCur[yy][xx] >= WHITE_PAWN) && (piecesValCur[yy][xx] <= WHITE_KING)))
  {
    pathY[count] = yy - 2; pathX[count] = xx - 2; pathVal[count] = piecesValCur[yy][xx]; count++;
  }
}

// *************************************************************************************************************************
// Get the legal positions a piece can move to
int getPaths(byte val, int x, int y, int turn)
{
  // This function starts with a position, and the piece (val) in that position, and gets the legal paths (the squares 
  // where that piece can legally move) for that piece. The path is stored in arrays  pathX[30], pathY[30], pathVal[30], 
  // and pathCount contains number of squares stored in the arrays. Two different methods are used. For the pawns, knights 
  // and kings a brute force method, where the legal moves are stored in arrays, or scanned in the case of the king. For the 
  // bishops, rooks and queens we use the function getSinglePathinPaths() to follow each line of squares where the piece can 
  // legally move. For the bishops, rooks and queens, each legal direction is placed in the calls to getSinglePathinPaths().
  
  int xx, yy;
  // Pawn ************************************************************************************ Pawn Pawn Pawn Pawn Pawn 
  if ((val == BLACK_PAWN) || (val == WHITE_PAWN))
  {
    // First 2 are straight ahead moves
    int wPawnX[4] = { 0, 0,-1, 1};
    int wPawnY[4] = {-1,-2,-1,-1};
    int bPawnX[4] = { 0, 0,-1, 1};
    int bPawnY[4] = { 1, 2, 1, 1};
    
    for (int i = 0; i < 4; i++)
    {
      if (i == 1) continue;           // The test for the second square is in the i = 0 block
      if (i == 0)
      {
        if (((turn == WHITE) && (piecesValCur[y + 2 + wPawnY[i]][x + 2 + wPawnX[i]] == EMPTY)) ||
            ((turn == BLACK) && (piecesValCur[y + 2 + bPawnY[i]][x + 2 + bPawnX[i]] == EMPTY)))
        {
          if (turn == WHITE)
          {
            pathY[pathCount] = y + wPawnY[i]; pathX[pathCount] = x + wPawnX[i]; pathVal[pathCount] = piecesValCur[y + 2 + wPawnY[i]][x + 2 + wPawnX[i]]; pathCount++;
          }
          else
          {
            pathY[pathCount] = y + bPawnY[i]; pathX[pathCount] = x + bPawnX[i]; pathVal[pathCount] = piecesValCur[y + 2 + bPawnY[i]][x + 2 + bPawnX[i]]; pathCount++;
          }
          if (((piecesValCur[y + 2 + wPawnY[i + 1]][x + 2 + wPawnX[i + 1]] == EMPTY) && (y == 6)) ||    // pawn in rank 2, starting position
              ((piecesValCur[y + 2 + bPawnY[i + 1]][x + 2 + bPawnX[i + 1]] == EMPTY) && (y == 1)))      // pawn in rank 7, starting position
          {
              if (turn == WHITE)
              {
                pathY[pathCount] = y + wPawnY[i + 1]; pathX[pathCount] = x + wPawnX[i + 1]; pathVal[pathCount] = piecesValCur[y + 2 + wPawnY[i + 1]][x + 2 + wPawnX[i + 1]]; pathCount++;
              }
              else
              {
                pathY[pathCount] = y + bPawnY[i + 1]; pathX[pathCount] = x + bPawnX[i + 1]; pathVal[pathCount] = piecesValCur[y + 2 + bPawnY[i + 1]][x + 2 + bPawnX[i + 1]]; pathCount++;
              }
          }
        }
      }
      else // i = 2 or 3, can pawn take a piece ?
      {
        if (((turn == WHITE) && (piecesValCur[y + 2 + wPawnY[i]][x + 2 + wPawnX[i]] >= BLACK_PAWN) && (piecesValCur[y + 2 + wPawnY[i]][x + 2 + wPawnX[i]] <= BLACK_KING)) ||
            ((turn == BLACK) && (piecesValCur[y + 2 + bPawnY[i]][x + 2 + bPawnX[i]] >= WHITE_PAWN)))
        {
          if (turn == WHITE)
          {
            pathY[pathCount] = y + wPawnY[i]; pathX[pathCount] = x + wPawnX[i]; pathVal[pathCount] = piecesValCur[y + 2 + wPawnY[i]][x + 2 + wPawnX[i]]; pathCount++;
          }
          else
          {
            pathY[pathCount] = y + bPawnY[i]; pathX[pathCount] = x + bPawnX[i]; pathVal[pathCount] = piecesValCur[y + 2 + bPawnY[i]][x + 2 + bPawnX[i]]; pathCount++;
          }
        }
      }
    }
  }

  // Knight *************************************************************************** Knight Knight Knight Knight 
  if ((val == BLACK_KNIGHT) || (val == WHITE_KNIGHT))
  {
    // x and y are the starting positions
    int i[8] = {1, 2, -1, -2, -1, -2, 1, 2};
    int j[8] = {2, 1, -2, -1,  2,  1,-2,-1};
    // myDebug(x, y, "  x and y coming in");
    // Check all eight possible moves
    for (int k = 0; k < 8; k++)
    {
      xx = x + 2 + i[k]; yy = y + 2 + j[k];
      if (((turn == WHITE) && (piecesValCur[yy][xx] >= BLACK_PAWN) && (piecesValCur[yy][xx] <= BLACK_KING)) ||
          ((turn == BLACK) && (piecesValCur[yy][xx] >= WHITE_PAWN) && (piecesValCur[yy][xx] <= WHITE_KING)) ||
           (piecesValCur[yy][xx] == EMPTY))
      {
        pathY[pathCount] = yy - 2; pathX[pathCount] = xx - 2; pathVal[pathCount] = piecesValCur[yy][xx]; pathCount++;
      }
    }
  }

  // Bishop *************************************************************************** Bishop Bishop Bishop Bishop 
  if ((val == BLACK_BISHOP) || (val == WHITE_BISHOP))
  {
    getSinglePathinPaths(x + 2, y + 2, 1, 1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1,-1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 1,-1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1, 1, pathCount, turn);
  }

  // Rook ******************************************************************************* Rook Rook Rook Rook Rook 
  if ((val == BLACK_ROOK) || (val == WHITE_ROOK))
  {
    getSinglePathinPaths(x + 2, y + 2, 1, 0, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1, 0, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 0, 1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 0,-1, pathCount, turn);
  }

  // Queen ******************************************************************************* Queen Queen Queen Queen 
  if ((val == BLACK_QUEEN) || (val == WHITE_QUEEN))
  {
    // x and y are the starting positions
    getSinglePathinPaths(x + 2, y + 2, 1, 1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1,-1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 1,-1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1, 1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 1, 0, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2,-1, 0, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 0, 1, pathCount, turn);
    getSinglePathinPaths(x + 2, y + 2, 0,-1, pathCount, turn);
  }

  // King ********************************************************************************* King King King King King 
  if ((val == BLACK_KING) || (val == WHITE_KING))
  {
    //myDebug(x, y, "  x and y coming in");
    for (int i = -1; i < 2; i++)
    {
       xx = x + 2 + i;
      for (int j = -1; j < 2; j++)
      {
        yy = y + 2 + j;
        if (!((i == 0) && (j == 0)))
        {
         // myDebug(xx - 2, yy - 2, "Square for King path");
          if (((turn == WHITE) && (piecesValCur[yy][xx] >= BLACK_PAWN) && (piecesValCur[yy][xx] <= BLACK_KING)) ||
              ((turn == BLACK) && (piecesValCur[yy][xx] >= WHITE_PAWN) && (piecesValCur[yy][xx] <= WHITE_KING)) ||
               (piecesValCur[yy][xx] == EMPTY))
          {
            pathY[pathCount] = yy - 2; pathX[pathCount] = xx - 2; pathVal[pathCount] = piecesValCur[yy][xx]; pathCount++;
          }
        }
      }
    }
  }
  return pathCount;
}

// *************************************************************************************************************************
// Color the positions a piece can legally move to
void colorPath(int count, int turn)
{
  for (int i = 0; i < count; i++)
  {
    if (pathVal[i] == EMPTY) colorSquare(pathX[i], pathY[i], colors[GREEN], true);
    if ((turn == WHITE) && (pathVal[i] >= BLACK_PAWN) && (pathVal[i] <= BLACK_KING)) colorSquare(pathX[i], pathY[i], colors[YELLOW], true);
    if ((turn == BLACK) && (pathVal[i] >= WHITE_PAWN) && (pathVal[i] <= WHITE_KING)) colorSquare(pathX[i], pathY[i], colors[YELLOW], true);
  }
}

// *************************************************************************************************************************
// Read a row of 8 Hall Effect sensors
int readLine(int row, byte piecesTemp[][8])
{
  // Read each sensor in a row looking for a zero, which means a piece is there.
  // The 3144 Hall Effect sensors are open collector output and pull to ground when activated.
  // We invert the zero to place a one in our array if a piece is on a square.
  // After each read we add to count to keep a running total.
  int count = 0;
  piecesTemp[row][0] = !digitalRead(2);
  count += piecesTemp[row][0];
  piecesTemp[row][1] = !digitalRead(3);
  count += piecesTemp[row][1];
  piecesTemp[row][2] = !digitalRead(4);
  count += piecesTemp[row][2];
  piecesTemp[row][3] = !digitalRead(5);
  count += piecesTemp[row][3];
  piecesTemp[row][4] = !digitalRead(6);
  count += piecesTemp[row][4];
  piecesTemp[row][5] = !digitalRead(7);
  count += piecesTemp[row][5];
  piecesTemp[row][6] = !digitalRead(8);
  count += piecesTemp[row][6];
  piecesTemp[row][7] = !digitalRead(12);  // Oddball because we skipped 9, 10 
  count += piecesTemp[row][7];            // for serial RX, TX, Serial2 on Teensy 3.1
  return count;                           // return number of pieces found in this row
}

// *************************************************************************************************************************
// Read all 8 lines of the Hall Effect Sensors
int readHall(byte piecesTemp[][8])
{
  // The function calls readLine() 8 times to read all 8 rows of Hall Effect sensors
  // When finished piecesTemp[8][8] has positions of pieces on board
  // Before each read we set the 3 bit address to the 74138
  // The total number of pieces on the board is returned
  int count = 0;
  digitalWrite(A0, LOW);
  digitalWrite(A1, LOW);
  digitalWrite(A2, LOW);    // 000
  delay(1);                 // let 3144 settle
  count += readLine(0, piecesTemp);
  digitalWrite(A0, HIGH);   // 001
  delay(1);
  count += readLine(1, piecesTemp);
  digitalWrite(A0, LOW);    // 010
  digitalWrite(A1, HIGH);
  delay(1);
  count += readLine(2, piecesTemp);
  digitalWrite(A0, HIGH);   // 011
  delay(1);
  count += readLine(3, piecesTemp);
  digitalWrite(A0, LOW);    // 100
  digitalWrite(A1, LOW);
  digitalWrite(A2, HIGH);
  delay(1);
  count += readLine(4, piecesTemp);
  digitalWrite(A0, HIGH);   // 101
  delay(1);
  count += readLine(5, piecesTemp);
  digitalWrite(A0, LOW);    // 110
  digitalWrite(A1, HIGH);
  delay(1);
  count += readLine(6, piecesTemp);
  digitalWrite(A0, HIGH);   // 111
  delay(1);
  count += readLine(7, piecesTemp);
  delay(100);
  return count;             // return count of all pieces on the board
}

// ************************************************************************************************************************
// Store or retrieve prievous board settings to be able to undo a move
void store_retrieveHist(int dir, byte piecesValCur[][12], byte piecesValHist[][12], byte piecesCurrent[][8], byte piecesHist[][8], byte bdCount, byte bdCountHist)
{
  // This is just a brute force store or retrieve of all the piece positions and values on the board.
  // The borders could have been eliminated.
  for (int y = 0; y < 12; y++)
    {
      for (int x = 0; x < 12; x++)
      {
        if (dir == STORE)
        {
          piecesValHist[y][x] = piecesValCur[y][x];
        }
        else if (dir == RETRIEVE)
        {
          piecesValCur[y][x] = piecesValHist[y][x];
        }
      }
    }

    for (int y = 0; y < 8; y++)
    {
      for (int x = 0; x < 8; x++)
      {
        if (dir == STORE)
        {
          piecesHist[y][x] = piecesCurrent[y][x];
        }
        else if (dir == RETRIEVE)
        {
          piecesCurrent[y][x] = piecesHist[y][x];
        }
      }
    }

    if (dir == STORE)
    {
      bdCountHist = bdCount;
    }
    else if (dir == RETRIEVE)
    {
      bdCount = bdCountHist;
    }
}

// *************************************************************************************************************************
// Color an x,y square with a specific color
void colorSquare(byte x, byte y, uint32_t color, bool on)
{
  if ((x >= 0) && (x < 8) && (y >= 0) && (y < 8)) // is it a legal square ?
  {
    int num1 = (y << 5) + (x << 1);         // same as num1 = y * 32 + x * 2;
    int num2 = num1 + 1;
    int num3 = (y << 5) + 31 - (x << 1);    // same as num3 = (y * 32 + 32) - (x * 2) - 1;
    int num4 = num3 - 1; 
    // Needs a short delay, otherwise intermittent errors on position
    pixels.setPixelColor(num1, color); delay(1);
    pixels.setPixelColor(num2, color); delay(1);
    pixels.setPixelColor(num3, color); delay(1);
    pixels.setPixelColor(num4, color); delay(1);
    if (on) pixels.show();  
  }
}

// *************************************************************************************************************************
// Color all squares, red for error alert, blue for checkmate
// Seems to work OK without a delay
void colorBoard(uint32_t color)
{
  for (int i = 0; i < NUMPIXELS; i++)
  {
    pixels.setPixelColor(i, color); 
  }
  pixels.show();
}

// *************************************************************************************************************************
// Color a chessboard with blue (for black) and red squares
void standBoard()
{
  for (int y = 0; y < 4; y++)
  {
    for (int x = 0; x < 4; x++)
    {
      colorSquare(x * 2, y * 2, colors[RED], false);
      delay(1);
      colorSquare(x * 2 + 1, y * 2, colors[BLUE], false);
      delay(1);
      colorSquare(x * 2, y * 2 + 1, colors[BLUE], false);
      delay(1);
      colorSquare(x * 2 + 1, y * 2 + 1, colors[RED], false);
      delay(1);
    }
  }
  pixels.show();
}

// *************************************************************************************************************************
// Show an error state and wait for correction, ie. all pieces replaced to positions before the error
void showError(byte piecesCurrent[][8], byte piecesError[][8])
{
  bool flag;
  int total;
  colorBoard(colors[RED]);
  do
  {
    flag = true;
    total = readHall(piecesError);
    for (int y = 0; y < 8; y++)
    {
      for (int x = 0; x < 8; x++)
      {
        if (piecesError[y][x] != piecesCurrent[y][x])
        {
          myDebug(x, y, "Error here");
          delay(1000);
          flag = false;
          break;
        }
        if (flag == false) break;   // needed for double loop
      }
    }
  } while (flag == false);
  standBoard();
}

// *************************************************************************************************************************
// The name of the piece for LCD readout and debugging
String namePiece(int x, int y)
{
  switch (piecesValCur[y + 2][x + 2])
  {
    case 0: return "Empty ";
            break;
    case 1: return "Black Pawn ";
            break;
    case 3: return "Black Knight ";
            break;
    case 4: return "Black Bishop ";
            break;
    case 5: return "Black Rook ";
            break;
    case 9: return "Black Queen ";
            break;
    case 16: return "Black King ";
            break;
    case 129: return "White Pawn ";
            break;
    case 131: return "White Knight ";
            break;
    case 132: return "White Bishop ";
            break;
    case 133: return "White Rook ";
            break;
    case 137: return "White Queen ";
            break;
    case 144: return "White King ";
            break;
  }
  return "ERROR";
}

// *************************************************************************************************************************
// print the piece locations to the Serial monitor for debugging
void printPiecesLoc(byte pieces[][8])
{
  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      // The if statements space things nicely
      if (pieces[y][x] < 10) Serial.print("  ");
      if ((pieces[y][x] >= 10) && (pieces[y][x] < 100)) Serial.print(" ");
      Serial.print(pieces[y][x]);
      Serial.print("  ");
    }
    Serial.println();
    Serial.println();
  }
  Serial.println();
  Serial.println();
}

// *************************************************************************************************************************
// print the piece values to the Serial monitor for debugging
void printPiecesVal(byte pieces[][12])
{
  for (int y = 2; y < 10; y++)
  {
    for (int x = 2; x < 10; x++)
    {
      // The if statements space things nicely
      if (pieces[y][x] < 10) Serial.print("  ");
      if ((pieces[y][x] >= 10) && (pieces[y][x] < 100)) Serial.print(" ");
      Serial.print(pieces[y][x]);
      Serial.print("  ");
    }
    Serial.println();
    Serial.println();
  }
  Serial.println();
  Serial.println();
}

// ************************************************************************************************************************
// Print the paths to the Serial monitor for debugging
void printPaths(int x, int y)
{
  Serial.print("Paths for ");
  Serial.println(namePiece(x, y));
  for (int i = 0; i < pathCount; i++)
  {
    Serial.print(pathX[i]);
    Serial.print("   ");
    Serial.print(pathY[i]);
    Serial.print("   ");
    Serial.print(pathVal[i]);
    Serial.println();
  }
  Serial.println("***********");
}

// ************************************************************************************************************************
// myDebug, print two numbers and a string to the Serial monitor for debugging
void myDebug(int x, int y, String mess)
{
  Serial.print(x);
  Serial.print("   ");
  Serial.print(y);
  Serial.print("   ");
  Serial.println(mess);
}

// ************************************************************************************************************************
// Return the traditional rank and file notation denoting the position of a square, so position 3,4 equals d5
// This is used for output to the LCD display
String rankFile(int x, int y)
{
  // This turns an x,y position into standard chess rank file notation and returns a string
  // x = columns, files, a in LL, a b c d e f g h
  // y = rows, ranks, 1 in LL, 1 2 3 4 5 6 7 8 
  String message = files[x] + ranks[y];
  return message;
}


